@@ -6,11 +6,47 @@ use async_recursion::async_recursion;
 use async_trait::async_trait;
 use dashmap::DashSet;
 use globset::{Glob, GlobSet, GlobSetBuilder};
+use hyper::{Client, Method, Request};
+use hyper_tls::HttpsConnector;
+use serde::{Deserialize, Serialize};
 use tokio::{fs, process::Command};
 use toml_edit::{value, Document};
 
 use mol_core::prelude::*;
 
+fn remove_start_dot(dir: PathBuf) -> PathBuf {
+  if dir.starts_with("./") {
+    dir.iter().skip(1).collect()
+  } else {
+    dir
+  }
+}
+
+#[derive(Debug, Serialize, Deserialize)]
+struct CratesError {
+  detail: String,
+}
+
+#[derive(Debug, Serialize, Deserialize)]
+struct CratesVersion {
+  version: CratesVersionMetadata,
+}
+
+#[derive(Debug, Serialize, Deserialize)]
+struct CratesVersionMetadata {
+  #[serde(alias = "crate")]
+  name: String,
+  num: String,
+  yanked: bool,
+}
+
+#[derive(Debug, Serialize, Deserialize)]
+#[serde(untagged)]
+enum CratesResult<T, E = CratesError> {
+  Ok(T),
+  Err { errors: Vec<E> },
+}
+
 #[derive(Default)]
 pub struct Cargo;
 
@@ -20,9 +56,9 @@ impl Cargo {
     exists: Arc<DashSet<PathBuf>>,
     globs: GlobSet,
     entry: fs::DirEntry,
-  ) -> std::io::Result<Vec<Package<V>>> {
+  ) -> anyhow::Result<Vec<Package<V>>> {
     let mut result = Vec::new();
-    let entry_path = entry.path();
+    let entry_path = remove_start_dot(entry.path());
 
     if exists.contains(&entry_path) {
       return Ok(result);
@@ -36,16 +72,16 @@ impl Cargo {
 
     if let Ok(file_type) = entry.file_type().await {
       if file_type.is_dir() {
-        return Cargo::check_read_dir(exists, globs, fs::read_dir(&entry_path).await?).await;
+        return Cargo::check_read_dir(exists, globs, fs::read_dir(entry.path()).await?).await;
       }
 
       if file_type.is_symlink() {
-        let entry_path = fs::read_link(&entry_path).await?;
-        return Cargo::check_read_dir(exists, globs, fs::read_dir(&entry_path).await?).await;
+        let link_value = fs::read_link(entry.path()).await?;
+        return Cargo::check_read_dir(exists, globs, fs::read_dir(&link_value).await?).await;
       }
 
-      if globs.is_match(entry.path()) && file_type.is_file() && entry.file_name() == "Cargo.toml" {
-        result.extend(Cargo.read_package(&entry_path).await?);
+      if globs.is_match(entry_path) && file_type.is_file() && entry.file_name() == "Cargo.toml" {
+        result.extend(Cargo.read_package(entry.path()).await?);
       }
     }
 
@@ -57,7 +93,7 @@ impl Cargo {
     exists: Arc<DashSet<PathBuf>>,
     globs: GlobSet,
     mut current_dir: fs::ReadDir,
-  ) -> std::io::Result<Vec<Package<V>>> {
+  ) -> anyhow::Result<Vec<Package<V>>> {
     let mut handles = Vec::new();
 
     while let Some(entry) = current_dir.next_entry().await? {
@@ -83,14 +119,14 @@ impl Cargo {
     command: &str,
     crate_path: T,
     args: Vec<&str>,
-  ) -> std::io::Result<()> {
+  ) -> anyhow::Result<()> {
     if let Ok(canon_path) = dunce::canonicalize(crate_path) {
       Command::new("cargo")
         .current_dir(canon_path)
         .arg(command)
         .args(args)
         .spawn()
-        .expect("cargo command failed to start")
+        .expect("Cargo command failed to start")
         .wait()
         .await?;
     }
@@ -101,7 +137,7 @@ impl Cargo {
   async fn load_document<T: AsRef<Path>>(
     &self,
     crate_path: T,
-  ) -> std::io::Result<(PathBuf, Document)> {
+  ) -> anyhow::Result<(PathBuf, Document)> {
     let document = fs::read_to_string(&crate_path)
       .await?
       .parse::<Document>()
@@ -113,10 +149,14 @@ impl Cargo {
 
 #[async_trait]
 impl PackageManager for Cargo {
+  fn default_path() -> &'static str {
+    "Cargo.toml"
+  }
+
   async fn read_package<T: AsRef<Path> + Send + Sync, V: Versioned + Send + Sync + 'static>(
     &self,
     crate_path: T,
-  ) -> std::io::Result<Vec<Package<V>>> {
+  ) -> anyhow::Result<Vec<Package<V>>> {
     let mut result = Vec::new();
     let (crate_path, document) = self.load_document(crate_path).await?;
 
@@ -131,19 +171,21 @@ impl PackageManager for Cargo {
 
     let mut dependencies = Vec::new();
 
-    // TODO: support [dependencies.xyz] patterns
-    if let Some(deps) = document["dependencies"].as_table() {
-      for (key, value) in deps.iter() {
-        if value.is_str() {
-          dependencies.push((
-            key.to_owned(),
-            value.as_str().unwrap_or_default().to_owned(),
-          ));
-        } else if value.is_table() || value.is_inline_table() {
-          dependencies.push((
-            key.to_owned(),
-            value["version"].as_str().unwrap_or_default().to_owned(),
-          ));
+    if document.contains_key("dependencies") {
+      // TODO: support [dependencies.xyz] patterns
+      if let Some(deps) = document["dependencies"].as_table() {
+        for (key, value) in deps.iter() {
+          if value.is_str() {
+            dependencies.push((
+              key.to_owned(),
+              value.as_str().unwrap_or_default().to_owned(),
+            ));
+          } else if value.is_table() || value.is_inline_table() {
+            dependencies.push((
+              key.to_owned(),
+              value["version"].as_str().unwrap_or_default().to_owned(),
+            ));
+          }
         }
       }
     }
@@ -191,11 +233,50 @@ impl PackageManager for Cargo {
     Ok(result)
   }
 
+  async fn check_version<V: Versioned + Send + Sync + 'static>(
+    &self,
+    package: &Package<V>,
+  ) -> anyhow::Result<bool> {
+    let https = HttpsConnector::new();
+    let client = Client::builder().build::<_, hyper::Body>(https);
+
+    let request = Request::builder()
+      .method(Method::GET)
+      .uri(format!(
+        "https://crates.io/api/v1/crates/{}/{}",
+        package.name, package.version.value
+      ))
+      .header(
+        hyper::header::USER_AGENT,
+        format!(
+          "mol-cargo/{} (https://github.com/DmitryDodzin/mol)",
+          env!("CARGO_PKG_VERSION")
+        ),
+      )
+      .body(hyper::Body::empty())?;
+
+    let response = client.request(request).await?;
+
+    let bytes = hyper::body::to_bytes(response.into_body()).await?;
+
+    let crates_result = serde_json::from_slice::<CratesResult<CratesVersion>>(&bytes)?;
+
+    match crates_result {
+      CratesResult::Ok(val) => Ok(val.version.name == package.name),
+      CratesResult::Err { errors } => {
+        for error in errors {
+          println!("crates-error:\n{}", error.detail);
+        }
+        Ok(false)
+      }
+    }
+  }
+
   async fn run_build<T: AsRef<Path> + Send + Sync>(
     &self,
     crate_path: T,
     build_args: Vec<String>,
-  ) -> std::io::Result<()> {
+  ) -> anyhow::Result<()> {
     self
       .run_command(
         "build",
@@ -210,7 +291,7 @@ impl PackageManager for Cargo {
     crate_path: T,
     publish_args: Vec<String>,
     dry_run: bool,
-  ) -> std::io::Result<()> {
+  ) -> anyhow::Result<()> {
     let args = if dry_run {
       vec!["--dry-run"]
         .into_iter()
@@ -227,7 +308,7 @@ impl PackageManager for Cargo {
     &self,
     crate_path: T,
     version: &str,
-  ) -> std::io::Result<()> {
+  ) -> anyhow::Result<()> {
     let (crate_path, mut document) = self.load_document(crate_path).await?;
 
     if document.contains_key("package") {
@@ -244,7 +325,7 @@ impl PackageManager for Cargo {
     crate_path: T,
     name: &str,
     version: &str,
-  ) -> std::io::Result<()> {
+  ) -> anyhow::Result<()> {
     let (crate_path, mut document) = self.load_document(crate_path).await?;
 
     if document.contains_key("dependencies") {
